/*
DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                Version 2, December 2004

Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>

Everyone is permitted to copy and distribute verbatim or modified
copies of this license document, and changing it is allowed as long
as the name is changed.

        DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

0. You just DO WHAT THE FUCK YOU WANT TO.

*/

/*
A very simple GUI display for the Hokuyo laser

Press SPACEBAR to dump the laser readings to a text file
Press ESC to quit
*/

#include <GL/glut.h>
#include <GL/glu.h>
#include <urg/UrgCtrl.h>
#include <iostream>
#include <cstdio>
#include <fstream>
#include <vector>
#include <string>

using namespace std;
using namespace qrk;

// Change these values to your need
const float MAX_RANGE = 6.0; // metres, add a bit extra for displaying purpose
const char device[] = "/dev/ttyACM0";
void *font = GLUT_BITMAP_9_BY_15;

float PIXELS_PER_METRE; // assigned at run time
int SCREEN_W, SCREEN_H; // screen width/height
UrgCtrl urg; // Hokuyo class
bool saveTextFile = false;
int currentFile = 1;

// GLUT functions
void reshape(int w, int h);
void draw();
void idle();
void keyboard(unsigned char key, int x, int y);
void renderBitmapString(float x, float y, float z, void *font, const char *string);

// Hokuyo functions
void initLaser();

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutInitWindowSize(600,600);
    glutCreateWindow("Hokuyo");

	glutDisplayFunc(draw);
    glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);

    initLaser();

	glutMainLoop();
}

void draw()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw the axis and labels
    glColor3f(1,1,0);

    // X axis
    glBegin(GL_LINES);
    glVertex2f(0, SCREEN_H/2);
    glVertex2f(SCREEN_W, SCREEN_H/2);
    glEnd();

    // Y axis
    glBegin(GL_LINES);
    glVertex2f(SCREEN_W/2, 0);
    glVertex2f(SCREEN_W/2, SCREEN_H);
    glEnd();

    // X axis label
    char xlabel[128];
    sprintf(xlabel, "%.2fm", (SCREEN_W/2) / PIXELS_PER_METRE);
    renderBitmapString(SCREEN_W-50, SCREEN_H/2-15, 0, font, xlabel);

    // Y axis label
    char ylabel[128];
    sprintf(ylabel, "%.2fm", (SCREEN_H/2) / PIXELS_PER_METRE);
    renderBitmapString(SCREEN_W/2 + 10, SCREEN_H-15, 0, font, ylabel);

    // Get a single scan from the laser
    vector <long> data;
    long timestamp;

    int n = urg.capture(data, &timestamp);

    ofstream textFile;

    // Optional save to file
    if(saveTextFile) {
        char filename[256];

        sprintf(filename, "data-%03d.txt", currentFile);
        currentFile++;

        textFile.open(filename, ios_base::trunc);

        if(textFile.bad()) {
            cerr << "Error opening " << filename << " for writing" << endl;
            saveTextFile = false;
        }
    }

    for (int i=0; i<n; i++) {
        float angle = urg.index2rad(i);
        long range = data[i]; // in millimetres

        if(range < urg.minDistance() || range > urg.maxDistance())
            continue;

        // Conver range to metres
        float rangeMetres = range*0.001;

        float xpos = SCREEN_W/2 + rangeMetres*cos(angle)*PIXELS_PER_METRE;
        float ypos = SCREEN_H/2 + rangeMetres*sin(angle)*PIXELS_PER_METRE;

        // angle range x y
        if(saveTextFile) {
            textFile << (angle*180/M_PI) << " " << range << " " << (range*cos(angle)) << " " << (rangeMetres*sin(angle)) << endl;
        }

        // Now draw the laser reading from the centre of the screen
        glColor3f(0,1,0);

        glPointSize(3.0);
        glBegin(GL_POINTS);
        glVertex2f(xpos,ypos);
        glEnd();
    }

    if(saveTextFile) {
        textFile.close();
        saveTextFile = false;
    }

    glutSwapBuffers();
}


void reshape(int w, int h)
{
    // Reset the coordinate system before modifying
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    // Set the viewport to be the entire window
    glViewport(0, 0, w, h);

    // Set to 2D rendering
    gluOrtho2D(0, w, 0, h);

    if(w > h)
        PIXELS_PER_METRE = w*0.5 / MAX_RANGE;
    else
        PIXELS_PER_METRE = h*0.5 / MAX_RANGE;

    SCREEN_W = w;
    SCREEN_H = h;
}

void idle()
{
    draw();
}

void initLaser()
{
    if (! urg.connect(device)) {
        printf("UrgCtrl::connect: %s\n", urg.what());
        exit(1);
    }
}

// http://www.lighthouse3d.com/opengl/glut/index.php?bmpfont
void renderBitmapString(float x, float y, float z, void *font, const char *string)
{
    const char *c;
    glRasterPos3f(x, y,z);

    for (c=string; *c != '\0'; c++) {
        glutBitmapCharacter(font, *c);
    }
}

void keyboard(unsigned char key, int x, int y)
{
    switch(key) {
        case ' ': // spacebar
        saveTextFile = true;
        break;

        case 27: // ESC
        exit(0);
        break;
    }
}


